
   /***************\
   *   rcsanon.c   *
   \***************/

#ifndef PROGNAME
#define PROGNAME "rcsanon"
#endif

#ifndef PREFIX
#define PREFIX "/opt/" PROGNAME
#endif

#define ACTION_FWD 0x00
#define ACTION_TAG 0x01

#define X_ERR(x) XXX_ERR(x)
#define XXX_ERR(x) "[ERROR] (line: " #x ") " 

#define _GNU_SOURCE
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <grp.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

struct params {
   in_port_t port;
   uid_t uid;
   gid_t gid;
   char *chrootdir;
   char *configfile;
   char *pidfile;
} prm = {
   4444,
   0,
   0,
   NULL,
   PREFIX "/etc/" PROGNAME ".conf",
   PREFIX "/tmp/" PROGNAME ".pid",
};

struct config {
   pid_t pid;
   unsigned short int action;
   char localhost[128];
   in_port_t localport;
   char remotehost[128];
   in_port_t remoteport;
   char filterhost[128];
   char signature[128];
   struct config *next;
} *cfg = NULL;

int stop = 0;

int main(int argc, char *argv[]);
void showusage(int exitcode);
void parseparams(int argc, char *argv[]);
void parseconfig(void);
void dopipe(struct config *cp);
void sighandler(const int signum);
void logme(int priority, const char *format, ...);

int rcsnet_main(in_port_t port);

int main(int argc, char *argv[])
{
   struct config *cp;
   pid_t pid;
   FILE *fp;
   int fd;
   struct sigaction sa;
   FILE *errlog;

   if(getpid() != getpgid(0)) {
      if(setsid() == -1) {
         logme(LOG_ERR, X_ERR(__LINE__) "setsid(): %s", strerror(errno)); 
         exit(EXIT_FAILURE);
      }
   }

   if(chdir("/")) {
      logme(LOG_ERR, X_ERR(__LINE__) "chdir(): %s", strerror(errno)); 
      exit(EXIT_FAILURE);
   }

   parseparams(argc, argv);

   if((fd = open("/dev/null", O_RDWR)) == -1) {
      logme(LOG_ERR, X_ERR(__LINE__) "open(): %s", strerror(errno)); 
      exit(EXIT_FAILURE);
   }
   if(dup2(fd, STDIN_FILENO) == -1) {
      logme(LOG_ERR, X_ERR(__LINE__) "dup2(): %s", strerror(errno)); 
      exit(EXIT_FAILURE);
   }
   if(dup2(fd, STDOUT_FILENO) == -1) {
      logme(LOG_ERR, X_ERR(__LINE__) "dup2(): %s", strerror(errno)); 
      exit(EXIT_FAILURE);
   }
   if(dup2(fd, STDERR_FILENO) == -1) {
      logme(LOG_ERR, X_ERR(__LINE__) "dup2(): %s", strerror(errno)); 
      exit(EXIT_FAILURE);
   }
   close(fd);

   sa.sa_handler = sighandler;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags = 0;
   if(sigaction(SIGHUP, &sa, NULL)) {
      logme(LOG_ERR, X_ERR(__LINE__) "sigaction(): %s", strerror(errno));
      exit(EXIT_FAILURE);
   }
   if(sigaction(SIGCHLD, &sa, NULL)) {
      logme(LOG_ERR, X_ERR(__LINE__) "sigaction(): %s", strerror(errno));
      exit(EXIT_FAILURE);
   }

   pid = getpid();

   if((fp = fopen(prm.pidfile, "w")) == NULL) {
      logme(LOG_ERR, X_ERR(__LINE__) "fopen(): %s", strerror(errno));
      exit(EXIT_FAILURE);
   }

   if(fprintf(fp, "%u", pid) < 0) {
      logme(LOG_ERR, X_ERR(__LINE__) "fprintf(): %s", strerror(errno));
      exit(EXIT_FAILURE);
   }

   fclose(fp);

   logme(LOG_INFO, "Daemon started (pid: %u)", pid);

   if(!(pid = fork())) {
      signal(SIGHUP, SIG_IGN);
      rcsnet_main(prm.port);
   }

   if(waitpid(pid, &fd, WNOHANG)) {
      exit(EXIT_FAILURE);
   }

   while(1) {
      unlink("/opt/rcsanon/tmp/errlog");

      parseconfig();

      for(cp = cfg; cp; cp = cp->next) {
         if((pid = fork()) == -1) {
            logme(LOG_ERR, X_ERR(__LINE__) "fork(): %s", strerror(errno)); 
            exit(EXIT_FAILURE);
         } else if(!pid) {
            sa.sa_handler = SIG_IGN;
            sigemptyset(&sa.sa_mask);
            sa.sa_flags = 0;
            if(sigaction(SIGHUP, &sa, NULL)) {
               logme(LOG_ERR, X_ERR(__LINE__) "sigaction(): %s", strerror(errno));
               exit(EXIT_FAILURE);
            }
            if(sigaction(SIGCHLD, &sa, NULL)) {
               logme(LOG_ERR, X_ERR(__LINE__) "sigaction(): %s", strerror(errno));
               exit(EXIT_FAILURE);
            }
            dopipe(cp);
         }

         cp->pid = pid;
      }

      pause();

      sa.sa_handler = SIG_IGN;
      sigemptyset(&sa.sa_mask);
      sa.sa_flags = 0;
      if(sigaction(SIGCHLD, &sa, NULL)) {
         logme(LOG_ERR, X_ERR(__LINE__) "sigaction(): %s", strerror(errno));
         exit(EXIT_FAILURE);
      }

      while(cfg) {
         kill(cfg->pid, SIGTERM);
         waitpid(cfg->pid, NULL, 0);
         cp = cfg->next;
         free(cfg);
         cfg = cp;
      }

      sa.sa_handler = sighandler;
      sigemptyset(&sa.sa_mask);
      sa.sa_flags = 0;
      if(sigaction(SIGCHLD, &sa, NULL)) {
         logme(LOG_ERR, X_ERR(__LINE__) "sigaction(): %s", strerror(errno));
         exit(EXIT_FAILURE);
      }

      if(stop) {
         if((errlog = fopen("/opt/rcsanon/tmp/errlog", "w")) != NULL) {
            fprintf(errlog, "Unable to run properly, stopped until reload");
            fclose(errlog);
         }
         logme(LOG_INFO, "Unable to run properly, stopped until reload");
         system("killall rcsanon");
         exit(0);
      }
   }
}

void showusage(int exitcode)
{
   printf("usage: " PROGNAME " [-u <user>] [-g <group>] [-d <chrootdir>] [-c <configfile>]\n");
   printf("       " PROGNAME " -h\n");

   exit(exitcode);
}

void parseparams(int argc, char *argv[])
{
   int opt;
   char *p;
   struct passwd *pw;
   struct group *gr;
   struct stat st;

   while((opt = getopt(argc, argv, "p:u:g:d:c:h")) != -1) {
      switch(opt) {
         case 'p':
            prm.port = (in_port_t)atoi(optarg);
            break;
         case 'u':
            for(p = optarg; isdigit(*p); p++);
            if(!*p) {
               prm.uid = atoi(optarg);
            } else if((pw = getpwnam(optarg))) {
               prm.uid = pw->pw_uid;
            } else {
               fprintf(stderr, "Invalid user: %s\n", optarg);
               exit(EXIT_FAILURE);
            }
            break;
         case 'g':
            for(p = optarg; isdigit(*p); p++);
            if(!*p) {
               prm.gid = atoi(optarg);
            } else if((gr = getgrnam(optarg))) {
               prm.gid = gr->gr_gid;
            } else {
               fprintf(stderr, "Invalid group: %s\n", optarg);
               exit(EXIT_FAILURE);
            }
            break;
         case 'd':
            if(stat(optarg, &st) || !S_ISDIR(st.st_mode)) {
               fprintf(stderr, "Invalid chrootdir: %s\n", optarg);
               exit(EXIT_FAILURE);
            }
            prm.chrootdir = optarg;
            break;
         case 'c':
            prm.configfile = optarg;
            break;
         case 'h':
            showusage(EXIT_SUCCESS);
         default:
            showusage(EXIT_FAILURE);
         }
   }

   if(stat(prm.configfile, &st) || !S_ISREG(st.st_mode)) {
      fprintf(stderr, "Unable to read configfile: %s\n", prm.configfile);
      exit(EXIT_FAILURE);
   }

   return;
}

void parseconfig(void)
{
   FILE *fp;
   char line[1024], *lp;
   struct config *cp;
   char action[4], localhost[128], remotehost[128], filterhost[128], signature[128];
   in_port_t localport, remoteport;

   logme(LOG_INFO, "Loading configuration using %s", prm.configfile);

   if((fp = fopen(prm.configfile, "r")) == NULL) {
      logme(LOG_ERR, X_ERR(__LINE__) "fopen(): %s", strerror(errno));
      exit(EXIT_FAILURE);
   }

   while(fgets(line, sizeof(line), fp)) {
      for(lp = line; isspace(*lp); lp++);
      if(!*lp || (*lp == '#') || (*lp == ';')) continue;

      if(sscanf(line, "%3s | %127s %hu -> %127s %hu | %127s | %127s ", action, localhost, &localport, remotehost, &remoteport, filterhost, signature) != 7) {
         logme(LOG_ERR, "[ERROR] Invalid config line: %s", line);
         exit(EXIT_FAILURE);
      }

      if(strcmp(remotehost, "0.0.0.0")) {
         logme(LOG_INFO, "  adding rule (local %s %hu, remote %s %hu)", localhost, localport, remotehost, remoteport);
      } else {
         logme(LOG_INFO, "  dropping rule (local %s %hu, remote %s %hu)", localhost, localport, remotehost, remoteport);
         continue;
      }

      if(!strcasecmp(action, "fwd")) {
         action[0] = ACTION_FWD;
      } else if(!strcasecmp(action, "tag")) {
         action[0] = ACTION_TAG;
      } else {
         logme(LOG_ERR, "[ERROR] Invalid config line: %s", line);
         exit(EXIT_FAILURE);
      }

      if(!cfg) {
         if(!(cfg = malloc(sizeof(struct config)))) {
            logme(LOG_ERR, X_ERR(__LINE__) "malloc(): %s", strerror(errno));
            exit(EXIT_FAILURE);
         }
         cp = cfg;
      } else {
         if(!(cp->next = malloc(sizeof(struct config)))) {
            logme(LOG_ERR, X_ERR(__LINE__) "malloc(): %s", strerror(errno));
            exit(EXIT_FAILURE);
         }
         cp = cp->next;
      }

      cp->action = (unsigned short int)(action[0]);
      memcpy(cp->localhost, localhost, sizeof(cp->localhost));
      cp->localport = localport;
      memcpy(cp->remotehost, remotehost, sizeof(cp->remotehost));
      cp->remoteport = remoteport;
      memcpy(cp->filterhost, filterhost, sizeof(cp->filterhost));
      memcpy(cp->signature, signature, sizeof(cp->signature));
      cp->next = NULL;
   }
 
   fclose(fp);

   if(cfg) {
      logme(LOG_INFO, "Configuration loaded");
      for(cp = cfg; cp; cp = cp->next) {
         logme(LOG_INFO, "> local %s %hu, remote %s %hu", cp->localhost, cp->localport, cp->remotehost, cp->remoteport);
         logme(LOG_INFO, "  action %s, filter %s", ((cp->action == ACTION_FWD) ? "fwd" : "tag"), cp->filterhost);
      }
   } else {
      logme(LOG_INFO, "Configuration is empty");
   }

   return;
}

void dopipe(struct config *cp)
{
   int bsock, lsock, rsock;
   struct sockaddr_in baddr, laddr, raddr;
   struct hostent *he;
   pid_t pid;
   char buf[4096];
   int len;
   fd_set rfds;
   FILE *errlog;
   FILE *connlog;

   if((bsock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
      logme(LOG_ERR, X_ERR(__LINE__) "socket(): %s", strerror(errno)); 
      exit(EXIT_FAILURE);
   }

   len = 1;
   if(setsockopt(bsock, SOL_SOCKET, SO_REUSEADDR, &len, sizeof(len))) {
      logme(LOG_ERR, X_ERR(__LINE__) "setsockopt(): %s", strerror(errno)); 
      exit(EXIT_FAILURE);
   }

   baddr.sin_family = AF_INET;
   baddr.sin_port = htons(cp->localport);

   if(!(he = gethostbyname(cp->localhost))) {
      logme(LOG_ERR, X_ERR(__LINE__) "gethostbyname(): unable to resolve %s", cp->localhost);
      exit(EXIT_FAILURE);
   } else {
      memcpy(&baddr.sin_addr.s_addr, he->h_addr_list[0], sizeof(baddr.sin_addr.s_addr));
   }

   if(bind(bsock, (struct sockaddr *)&baddr, sizeof(baddr))) {
      logme(LOG_ERR, X_ERR(__LINE__) "bind(): %s", strerror(errno));
      exit(EXIT_FAILURE);
   }

   if(listen(bsock, 10)) {
      logme(LOG_ERR, X_ERR(__LINE__) "listen(): %s", strerror(errno));
      exit(EXIT_FAILURE);
   }

   if(prm.chrootdir) {
      if(chroot(prm.chrootdir)) {
         logme(LOG_ERR, X_ERR(__LINE__) "chroot(): %s", strerror(errno));
         exit(EXIT_FAILURE);
      }

      if(chdir("/")) {
         logme(LOG_ERR, X_ERR(__LINE__) "chdir(): %s", strerror(errno));
         exit(EXIT_FAILURE);
      }
   }

   if(prm.gid) {
      if(setgroups(0, NULL)) {
         logme(LOG_ERR, X_ERR(__LINE__) "setgroups(): %s", strerror(errno));
         exit(EXIT_FAILURE);
      }

      if(setgid(prm.gid)) {
         logme(LOG_ERR, X_ERR(__LINE__) "setgid(): %s", strerror(errno));
         exit(EXIT_FAILURE);
      }
   }

   if(prm.uid) {
      if(setuid(prm.uid)) {
         logme(LOG_ERR, X_ERR(__LINE__) "setuid(): %s", strerror(errno));
         exit(EXIT_FAILURE);
      }
   }

   logme(LOG_INFO, "Listening on %s %hu", cp->localhost, cp->localport);

   while(1) {
      len = sizeof(laddr);
      if((lsock = accept(bsock, (struct sockaddr *)&laddr, (unsigned int *)&len)) == -1) {
         logme(LOG_ERR, X_ERR(__LINE__) "accept(): %s", strerror(errno));
         continue;
      }

      if(strcmp(cp->filterhost, "*")) {
         if((he = gethostbyname(cp->filterhost)) && memcmp(&laddr.sin_addr.s_addr, he->h_addr_list[0], sizeof(laddr.sin_addr.s_addr))) {
            logme(LOG_INFO, "Denied connection from %s %hu", inet_ntoa(laddr.sin_addr), ntohs(laddr.sin_port));
            close(lsock);

            if((connlog = fopen("/opt/rcsanon/tmp/connlog", "a")) != NULL) {
               fprintf(connlog, "Denied connection from %s %hu\n", inet_ntoa(laddr.sin_addr), ntohs(laddr.sin_port));
               fclose(connlog);
            }

            continue;
         }
      }

      logme(LOG_INFO, "Accepted connection from %s %hu", inet_ntoa(laddr.sin_addr), ntohs(laddr.sin_port));
      if((connlog = fopen("/opt/rcsanon/tmp/connlog", "a")) != NULL) {
         fprintf(connlog, "Accepted connection from %s %hu\n", inet_ntoa(laddr.sin_addr), ntohs(laddr.sin_port));
         fclose(connlog);
      }

      if((pid = fork()) == -1) {
         logme(LOG_ERR, X_ERR(__LINE__) "fork(): %s", strerror(errno));
         continue;
      } else if(pid) {
         close(lsock);
         continue;
      }

      close(bsock);

      if((rsock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
         logme(LOG_ERR, X_ERR(__LINE__) "socket(): %s", strerror(errno));
         exit(EXIT_FAILURE);
      }

      raddr.sin_family = AF_INET;
      raddr.sin_port = htons(cp->remoteport);

      if(!(he = gethostbyname(cp->remotehost))) {
         if((errlog = fopen("/opt/rcsanon/tmp/errlog", "w")) != NULL) {
            fprintf(errlog, "Error resolving %s", cp->remotehost);
            fclose(errlog);
         }
         logme(LOG_ERR, X_ERR(__LINE__) "gethostbyname(): unable to resolve %s", cp->remotehost);
         exit(EXIT_FAILURE);
      } else {
         memcpy(&raddr.sin_addr.s_addr, he->h_addr_list[0], sizeof(raddr.sin_addr.s_addr));
         unlink("/opt/rcsanon/tmp/errlog");
      }

      if(connect(rsock, (struct sockaddr *)&raddr, sizeof(raddr))) {
         if((errlog = fopen("/opt/rcsanon/tmp/errlog", "w")) != NULL) {
            fprintf(errlog, "Error connecting to %s", cp->remotehost);
            fclose(errlog);
         }
         logme(LOG_ERR, X_ERR(__LINE__) "connect(): %s", strerror(errno));
         exit(EXIT_FAILURE);
      } else {
         unlink("/opt/rcsanon/tmp/errlog");
      }

      if(cp->action == ACTION_TAG) {
         if(send(rsock, cp->signature, strlen(cp->signature), 0) != (strlen(cp->signature))) {
            logme(LOG_ERR, X_ERR(__LINE__) "send(): %s", strerror(errno));
            exit(EXIT_FAILURE);
         }

         if(send(rsock, (void *)&laddr.sin_addr.s_addr, sizeof(laddr.sin_addr.s_addr), 0) != sizeof(laddr.sin_addr.s_addr)) {
            logme(LOG_ERR, X_ERR(__LINE__) "send(): %s", strerror(errno));
            exit(EXIT_FAILURE);
         }
      }

      while(1) {
         FD_ZERO(&rfds);
         FD_SET(lsock, &rfds);
         FD_SET(rsock, &rfds);

         if(select(((lsock > rsock) ? lsock + 1 : rsock + 1), &rfds, NULL, NULL, NULL) == -1) {
            logme(LOG_ERR, X_ERR(__LINE__) "select(): %s", strerror(errno));
            exit(EXIT_FAILURE);
         }

         if(FD_ISSET(rsock, &rfds)) {
            if((len = recv(rsock, buf, sizeof(buf), 0)) == -1) {
               logme(LOG_ERR, X_ERR(__LINE__) "recv(): %s", strerror(errno));
               exit(EXIT_FAILURE);
            } else if(!len) {
               close(rsock);
               close(lsock);
               break;
            }

            if(send(lsock, buf, len, 0) != len) {
               logme(LOG_ERR, X_ERR(__LINE__) "send(): %s", strerror(errno));
               exit(EXIT_FAILURE);
            }
         }

         if(FD_ISSET(lsock, &rfds)) {
            if((len = recv(lsock, buf, sizeof(buf), 0)) == -1) {
               logme(LOG_ERR, X_ERR(__LINE__) "recv(): %s", strerror(errno));
               exit(EXIT_FAILURE);
            } else if(!len) {
               close(lsock);
               close(rsock);
               break;
            }

            if(send(rsock, buf, len, 0) != len) {
               logme(LOG_ERR, X_ERR(__LINE__) "send(): %s", strerror(errno));
               exit(EXIT_FAILURE);
            }
         }
      }

      logme(LOG_INFO, "Closed connection with %s %hu", inet_ntoa(laddr.sin_addr), ntohs(laddr.sin_port));

      exit(EXIT_SUCCESS);
   }
}

void sighandler(const int signum)
{
   switch(signum) {
      case SIGHUP:
         break;
      case SIGCHLD:
         stop = 1;
         break;
   }

   return;
}

void logme(int priority, const char *format, ...)
{
   va_list ap;
   FILE *fp;

   va_start(ap, format);

   openlog(PROGNAME, LOG_PID, LOG_DAEMON);
   vsyslog(priority, format, ap);
   closelog();

   if((fp = fopen("/opt/rcsanon/tmp/logme", "a")) != NULL) {
      fprintf(fp, "%u ", (unsigned int)time(NULL));
      vfprintf(fp, format, ap);
      fprintf(fp, "\n");
      fclose(fp);
   }

   va_end(ap);

   return;
}
