/*
    MODULE -- network configuration module (RNC)

    Copyright (C) Alberto Ornaghi

    $Id: netconf.c 1137 2009-11-13 15:25:56Z alor $
*/

#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <strings.h>
#include <errno.h>
#include <openssl/ssl.h>
#include <openssl/bio.h>
#include <netdb.h>
#include <signal.h>
#include <sys/statvfs.h>
#include <syslog.h>
#include "rcsnet.h"

/* globals */

#define LOOP for(;;)
#define DEBUG_MSG(x, ...) logme(LOG_INFO, x "\n", ## __VA_ARGS__ )
#define ERROR_MSG(x, ...) do { logme(LOG_ERR, x, ## __VA_ARGS__ ); exit(-1); } while(0)
/*#define DEBUG_MSG(x, ...) printf(x "\n", ## __VA_ARGS__ )*/
/*#define ERROR_MSG(x, ...) do { printf(x, ## __VA_ARGS__ ); exit(-1); } while(0)*/
#define SAFE_CALLOC(x, n, s) do { \
   x = calloc(n, s); \
	if (x == NULL) \
   	ERROR_MSG("virtual memory exhausted"); \
} while(0)
#define SAFE_FREE(x) do{ if(x) { free(x); x = NULL; } }while(0)

/* protos */

static int tcp_accept(int sock);
int ssl_proto_read(SSL *ssl, void *buf, int num);
int ssl_proto_write(SSL *ssl, void *buf, int num);
void rnc_handleproto(SSL *ssl);
int rnc_sendversion(SSL *ssl);
int rnc_sendmonitor(SSL *ssl, char *status, char *desc);
int rnc_retrieveconf(SSL *ssl);
int rnc_sendlogs(SSL *ssl);
void get_system_stats(u_int *disk, u_int *cpu, u_int *pcpu);

char sign[RNC_SIGN_LEN];

/************************************************/


int rcsnet_main(in_port_t listenport);

int rcsnet_main(in_port_t listenport)
{
   struct sockaddr_in saddr;
   SSL_CTX *ctx;
   SSL *ssl;
   int sock, asock;
   int opt;
   char *certfile;
   FILE *signfile;   

   if(!(signfile = fopen("/opt/rcsanon/etc/network.sig", "r"))) {
      ERROR_MSG("Cannot load the signature from /opt/rcsanon/etc/network.sig");
   }

   if(fread(sign, RNC_SIGN_LEN, 1, signfile) != 1) {
      ERROR_MSG("Cannot load the signature from /opt/rcsanon/etc/network.sig");
   }

   fclose(signfile);

   DEBUG_MSG("Signature: %.32s", sign);

   SSL_library_init();
   SSL_load_error_strings();

   /* create the SSL stuff */
   ctx = SSL_CTX_new(SSLv23_server_method());

   certfile = "/opt/rcsanon/etc/ca.pem";

   if (SSL_CTX_use_certificate_file(ctx, certfile, SSL_FILETYPE_PEM) == 0)
      ERROR_MSG("Cannot load the certificate from %s", certfile);

   if (SSL_CTX_use_PrivateKey_file(ctx, certfile, SSL_FILETYPE_PEM) <= 0)
      ERROR_MSG("Cannot load the private key from %s", certfile);

   DEBUG_MSG("SSL_CTX initialized [%s]", certfile);

   /* create, bind and listen on port tcp/RNC_LISTEN_PORT */
   saddr.sin_family = AF_INET;
	/* XXX - listening port configurable */
   saddr.sin_port = htons(listenport);
   saddr.sin_addr.s_addr = htonl(0);

   if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == 0)
      ERROR_MSG("Cannot create socket on port [%d]", ntohs(saddr.sin_port));

   /* reuse the socket */
   opt = 1;
   setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));

   if ((bind(sock, (const struct sockaddr *) &saddr, sizeof(struct sockaddr))) < 0)
      ERROR_MSG("Cannot bind the socket");

   listen(sock, 10);

   DEBUG_MSG("Server listening on port %d", ntohs(saddr.sin_port));

   /* main loop waiting to be contacted by RNC */
   LOOP {
      asock = tcp_accept(sock);

      ssl = SSL_new(ctx);

      /* associate the socket with SSL */
      SSL_set_fd(ssl, asock);
      SSL_set_accept_state(ssl);

      /* start the SSL handshake */
      if (SSL_accept(ssl)) {

         /* handle the communication with the RNC */
         rnc_handleproto(ssl);
      }

      DEBUG_MSG("Closing connection");

      close(asock);
      SSL_free(ssl);

      sleep(10);
   }

   SSL_CTX_free(ctx);
   close(sock);

   /* NEVER REACHED */
   exit(0);
}


static int tcp_accept(int sock)
{
   struct sockaddr_in caddr;
   u_int len = sizeof(struct sockaddr);
   fd_set  fdread;
   struct timeval tv;
   int ret;
   int csock;

   FD_ZERO(&fdread);
   FD_SET(sock, &fdread);
   memset(&caddr, 0, sizeof(caddr));

   /* set the timeout */
   tv.tv_sec = 0;
   tv.tv_usec = 0;

   ret = select(FOPEN_MAX, &fdread, (fd_set *)NULL, (fd_set *)NULL, NULL /*&tv*/);

   if (ret == 0) {
      /* timeout occurred. return false to let the main thread do other things... */
      return 0;
   } else if (ret == -1) {
      DEBUG_MSG("tcp_accept - select socket error [%d]", errno);
      return 0;
   }

   csock = accept(sock, (struct sockaddr *)&caddr, &len);

   if (csock == -1 ) {
      DEBUG_MSG("tcp_accept - invalid socket [%d]", errno);
      return 0;
   } else {
      DEBUG_MSG("New connection from %s", inet_ntoa(caddr.sin_addr));
      return csock;
   }

   return 0;
}

int ssl_proto_read(SSL *ssl, void *buf, int num)
{
   int read = 0;
   int len;
   do {
      len = SSL_read(ssl, (char *)buf + read, num - read);
      if (len <= 0) {
         break;
      }
      read += len;
   } while (read < num);

   return read;
}

int ssl_proto_write(SSL *ssl, void *buf, int num)
{
   int written = 0;
   int len;

   do {
      len = SSL_write(ssl, (char *)buf + written, num - written);
      if (len <= 0) {
         break;
      }
      written += len;
   } while (written < num);

   return written;
}


void rnc_handleproto(SSL *ssl)
{
   RncProtoHeader pheader;
   RncProtoLogin plogin;
   int ret;
   char descr[1024];
   FILE *errlog;
   FILE *pidfile;
   pid_t pid;

   DEBUG_MSG("Handling connection from RNC");

   /* read the login from RNC */
   if ( (ret = ssl_proto_read(ssl, &pheader, sizeof(pheader))) <= 0) {
      DEBUG_MSG("Cannot read from RNC [0x%08x]", SSL_get_error(ssl, ret));
      return;
   }

   /* check if the command is correct */
   if (ret < (int)sizeof(RncProtoHeader) || pheader.code != RNC_PROTO_LOGIN) {
      DEBUG_MSG("Invalid login authentication");
      return;
   }

   /* check if the signature is correct. otherwise reply with NO */
   ret = ssl_proto_read(ssl, &plogin, sizeof(plogin));
   if (ret < (int)sizeof(RncProtoLogin) || memcmp(plogin.sign, sign, RNC_SIGN_LEN)) {
      DEBUG_MSG("Invalid RNC authentication [%s]", plogin.sign);
      pheader.code = RNC_PROTO_NO;
      pheader.size = 0;
      ssl_proto_write(ssl, &pheader, sizeof(pheader));
      return;
   }

   pheader.code = RNC_PROTO_OK;
   pheader.size = 0;
   if (ssl_proto_write(ssl, &pheader, sizeof(pheader)) <= 0) {
      DEBUG_MSG("Cannot write to RNC [0x%08x]", SSL_get_error(ssl, ret));
      return;
   }

   DEBUG_MSG("RNC authenticated and connected");

   if(rnc_sendversion(ssl) < 0) {
      DEBUG_MSG("Cannot communicate with RNC (version)");
      return;
   }

   /* prepare the string for the monitor */
   pidfile = NULL;
   if(!(pidfile = fopen("/opt/rcsanon/tmp/rcsanon.pid", "r"))) {
      rnc_sendmonitor(ssl, "KO", "rcsanon is not running");
   } else if(fscanf(pidfile, "%u", &pid) != 1) {
      rnc_sendmonitor(ssl, "KO", "rcsanon is not running");
   } else if(kill(pid, 0)) {
      rnc_sendmonitor(ssl, "KO", "rcsanon is not running");
   } else if((errlog = fopen("/opt/rcsanon/tmp/errlog", "r"))) {
      while(fgets(descr, sizeof(descr), errlog)) {
         rnc_sendmonitor(ssl, "KO", descr);
      }
      fclose(errlog);
   } else {
      rnc_sendmonitor(ssl, "OK", "Running");
   }

   if(pidfile) fclose(pidfile);

   /* retrieve new conf (if any) */
   if ((ret = rnc_retrieveconf(ssl)) < 0) {
      DEBUG_MSG("Cannot communicate with RNC (config)");
      return;
   }

   /* check if there are new configs */
   if (ret) {
      DEBUG_MSG("Received new configuration(s), sending signal to reload them...");

      /* reload the new config, the signal handler will reload them */
      system("killall -HUP rcsanon");

   } else {
      DEBUG_MSG("NO new configuration this time...");
   }

   /* send cached logs */
   if ((ret = rnc_sendlogs(ssl)) < 0) {
      DEBUG_MSG("Cannot communicate with RNC (logs)");
      return;
   }

   /* send BYE to RNC */
   pheader.code = RNC_PROTO_BYE;
   pheader.size = 0;

   ssl_proto_write(ssl, &pheader, sizeof(pheader));

   /* disconnect */
}

int rnc_sendversion(SSL *ssl)
{
   RncProtoHeader pheader;
   RncProtoVersion pversion;

   memset(&pheader, 0, sizeof(pheader));
   memset(&pversion, 0, sizeof(pversion));

   /* header parameters */
   pheader.code = RNC_PROTO_VERSION;
   pheader.size = sizeof(pversion);

   /* monitor parameters */
   snprintf(pversion.version, sizeof(pversion.version), "%u", VERSION);

   DEBUG_MSG("Sending version information to RNC [%u]", VERSION);

   /* send header */
   if (ssl_proto_write(ssl, &pheader, sizeof(pheader)) <= 0)
      return -1;

   /* monitor part */
   if (ssl_proto_write(ssl, &pversion, sizeof(pversion)) <= 0)
      return -1;

   /* read the response from RNC */
   if (ssl_proto_read(ssl, &pheader, sizeof(pheader)) <= 0)
      return -1;

   if (pheader.code != RNC_PROTO_OK)
      return -1;

   return 0;
}


int rnc_sendmonitor(SSL *ssl, char *status, char *desc)
{
   RncProtoHeader pheader;
   RncProtoMonitor pmonitor;

   memset(&pheader, 0, sizeof(pheader));
   memset(&pmonitor, 0, sizeof(pmonitor));

   /* header parameters */
   pheader.code = RNC_PROTO_MONITOR;
   pheader.size = sizeof(pmonitor);

   /* monitor parameters */
   snprintf(pmonitor.status, sizeof(pmonitor.status), "%s", status);
   get_system_stats(&pmonitor.disk, &pmonitor.cpu, &pmonitor.pcpu);
   snprintf(pmonitor.desc, sizeof(pmonitor.desc), "%s", desc);

   DEBUG_MSG("Sending monitor information to RNC [%s]", desc);

   /* send header */
   if (ssl_proto_write(ssl, &pheader, sizeof(pheader)) <= 0)
      return -1;

   /* monitor part */
   if (ssl_proto_write(ssl, &pmonitor, sizeof(pmonitor)) <= 0)
      return -1;

   /* read the response from RNC */
   if (ssl_proto_read(ssl, &pheader, sizeof(pheader)) <= 0)
      return -1;

   if (pheader.code != RNC_PROTO_OK)
      return -1;

   return 0;
}


int rnc_retrieveconf(SSL *ssl)
{
   FILE *fc;
   RncProtoHeader pheader;
   RncProtoConfig pconfig;
   int found = 0;
   char *conf;
   char path[1024];

   /* header parameters */
   pheader.code = RNC_PROTO_CONF;
   pheader.size = 0;

   /* send request to check if there is new config */
   if (ssl_proto_write(ssl, &pheader, sizeof(pheader)) <= 0)
      return -1;

   /* loop to receive the new conf */
   LOOP {
      memset(&pheader, 0, sizeof(pheader));
      memset(&pconfig, 0, sizeof(pconfig));

      /* read the response from RNC */
      if (ssl_proto_read(ssl, &pheader, sizeof(pheader)) <= 0)
         break;

      /* there is NOT a new config */
      if (pheader.code != RNC_PROTO_CONF)
         break;

      /* retrieve the config header */
      if (ssl_proto_read(ssl, &pconfig, sizeof(pconfig)) <= 0)
         break;

      /* allocate the buffer and read the conf from RNC */
      SAFE_CALLOC(conf, pconfig.size, sizeof(char));
      if (ssl_proto_read(ssl, conf, pconfig.size) <= 0)
         break;

      DEBUG_MSG("Received new config file [%s]", pconfig.filename);

      /* open the config file for writing */
      snprintf(path, sizeof(path), "/opt/rcsanon/tmp/%s", pconfig.filename);
      fc = fopen(path, "wb");
		if (fc == NULL) {
			ERROR_MSG("Cannot open %s", pconfig.filename);
		}

      /* dump the content of the buffer received from RNC into the file */
      if (fwrite(conf, sizeof(char), pconfig.size, fc) < pconfig.size)
         DEBUG_MSG("Cannot write conf file [%s]", pconfig.filename);

      DEBUG_MSG("Config file [%s] written (%d bytes)", pconfig.filename, pconfig.size);

      fclose(fc);

      /* if the file is a ZIP archive, extract it */
      if (!strcasecmp(pconfig.filename + strlen(pconfig.filename) - 4, ".zip")) {
         char *dir, *p;
         char argv[1024];
         int ret;

         dir = strdup(path);

         /* trim the filename, get the dirname */
         if ((p = strrchr(dir, '/')) != NULL)
            *p = 0;

         /* prepare the commandline for unzip */
         snprintf(argv, sizeof(argv), "unzip -o %s -d /opt/rcsanon/", path);

         DEBUG_MSG("Uncompressing configuration file...");
         /* execute the command */
         ret = system(argv);

         if (ret == -1 || ret == 127)
            DEBUG_MSG("Unzip failed");

         unlink(path);

         unlink("/opt/rcsanon/tmp/errlog");

         SAFE_FREE(dir);
      }

      /* increment the number of received config */
      found++;
   }

   return found;
}


int rnc_sendlogs(SSL *ssl)
{
   RncProtoHeader pheader;
   RncProtoLog plog;
   u_int count = 0;
   FILE *fp;
   char line[1024], *desc;
   time_t logtime;
   struct tm *logtm;

   /* header parameters */
   pheader.code = RNC_PROTO_LOG;
   pheader.size = sizeof(plog);

   if((fp = fopen("/opt/rcsanon/tmp/logme", "r"))) {
      while(fgets(line, sizeof(line), fp)) {
         if(sscanf(line, "%u", &logtime, desc) != 1) continue;

         desc = strchr(line, ' ');
         desc++;

         logtm = gmtime(&logtime);

         memset(&plog, 0, sizeof(plog));

         plog.timestamp.tm_sec = logtm->tm_sec;
         plog.timestamp.tm_min = logtm->tm_min;
         plog.timestamp.tm_hour = logtm->tm_hour;
         plog.timestamp.tm_mday = logtm->tm_mday;
         plog.timestamp.tm_mon = logtm->tm_mon + 1;
         plog.timestamp.tm_year = logtm->tm_year + 1900;
         plog.timestamp.tm_wday = logtm->tm_wday;
         plog.timestamp.tm_yday = logtm->tm_yday;
         plog.timestamp.tm_isdst = logtm->tm_isdst;

         plog.type = RNC_LOG_INFO;

         strncpy(plog.desc, desc, sizeof(plog.desc));

         /* send header for the log */

         if (ssl_proto_write(ssl, &pheader, sizeof(pheader)) <= 0)
            return -1;

         /* send the log */
         if (ssl_proto_write(ssl, &plog, sizeof(plog)) <= 0)
            return -1;

         //DEBUG_MSG("rnc_sendlogs - [%s]", plog.desc);

         count++;
      }
      fclose(fp);
      unlink("/opt/rcsanon/tmp/logme");
   }

   DEBUG_MSG("%d log sent to RNC", count);

   return count;
}


void get_system_stats(u_int *disk, u_int *cpu, u_int *pcpu)
{
   FILE *fproc;
   char line[1024];
   int ouser, onice, osys, oidle, ohi, oirq, osoft;
   int user, nice, sys, idle, hi, irq, soft;
   int opuser, opsys, puser, psys;
   int tot;
   char *p;
   struct statvfs fs;
   int dummy;
   char cdummy;

   /* initialize the values */
   *disk = -1;
   *cpu = -1;
   *pcpu = -1;

   /* filesystem stats */
   statvfs(".", &fs);
   *disk = (int)((float)fs.f_bavail / (float)fs.f_blocks * 100);

   memset(line, 0, sizeof(line));

   /* cpu stats (globals) */
   if ((fproc = fopen("/proc/stat", "r")) == NULL)
      return;
   dummy = fread(line, 1024 - 1, sizeof(char), fproc);
   fclose(fproc);
   /* get the values from the string (we need all of them) */
   if (sscanf(line, "cpu  %d %d %d %d %d %d %d", &ouser, &onice, &osys, &oidle, &ohi, &oirq, &osoft) != 7)
      return;

   memset(line, 0, sizeof(line));

   /* cpu stats (current process) */
   if ((fproc = fopen("/proc/self/stat", "r")) == NULL)
      return;
   dummy = fread(line, 1024 - 1, sizeof(char), fproc);
   fclose(fproc);

   /* skip the process name */
   if ((p = strchr(line, ')')) == NULL)
      return;

   /* get the values from the string (we need only user and sys times) */
   if (sscanf(p + 2, "%c %d %d %d %d %d %d %d %d %d %d %d %d",
         &cdummy, &dummy, &dummy,
         &dummy, &dummy, &dummy,
         &dummy, &dummy, &dummy,
         &dummy, &dummy,
         &opuser, &opsys) != 13)
      return;

   /* wait 1 second for the sampling */
   sleep(1);

   memset(line, 0, sizeof(line));

   if ((fproc = fopen("/proc/stat", "r")) == NULL)
      return;
   dummy = fread(line, 1024 - 1, sizeof(char), fproc);
   fclose(fproc);

   /* get the values from the string (we need all of them) */
   if (sscanf(line, "cpu  %d %d %d %d %d %d %d", &user, &nice, &sys, &idle, &hi, &irq, &soft) != 7)
      return;

   memset(line, 0, sizeof(line));

   if ((fproc = fopen("/proc/self/stat", "r")) == NULL)
      return;
   dummy = fread(line, 1024 - 1, sizeof(char), fproc);
   fclose(fproc);

   /* skip the process name */
   if ((p = strchr(line, ')')) == NULL)
      return;

   /* get the values from the string (we need only user and sys times) */
   if (sscanf(p + 2, "%c %d %d %d %d %d %d %d %d %d %d %d %d",
         &cdummy, &dummy, &dummy,
         &dummy, &dummy, &dummy,
         &dummy, &dummy, &dummy,
         &dummy, &dummy,
         &puser, &psys) != 13)
      return;

   tot = (user+nice+sys+idle+hi+irq+soft) - (ouser+onice+osys+oidle+ohi+oirq+osoft);

   *cpu = (int)((1 - (float)(idle - oidle) / (float)tot) * 100);
   if(*cpu < 0) *cpu = 0;
   if(*cpu > 100) *cpu = 0;
   *pcpu = (int)((float)(puser + psys - opuser - opsys) / (float)tot * 100);
   if(*pcpu < 0) *pcpu = 0;
   if(*pcpu > 100) *pcpu = 0;
}

/* EOF */

// vim:ts=3:expandtab

